PGDataLogger - логирование изменений данных в БД PostgreSQL
===========================================================

Основной целью данного логирования является необходимость связывания факта 
изменения данных с информацией о сессии, в рамках которой оно было сделано.

Т.е. необходимо, например, знать какой пользователь, каким запросом изменил 
определенную запись в базе данных.

Для этого, каждое изменение данных сопровождается информацией о сессии, в 
которой можно хранить любые необходимые данные.

```
2019-09-24 13:08:34,197 - pgdatalogger - INFO - {"_txid": 32869, "_kind": "BEGIN", "_session": "{\"id\": \"e2b1fb0617e8\", \"path\": \"accounting.registry_recalculate.tasks.recalculate_register\", \"user\": \"\", \"is_celery\": true}"} | {"kind": "insert", "table": "account_entry_cumreg", "key": null, "record": {"id": 16248974, "date": "2018-07-19 23:00:00+03", "date_next": "9999-12-30 23:00:00+03", "date_prev": "0001-12-31 23:09:57+02:30:17 BC", "ent_id": 42, "account_id": 459481, "kbk_id": 17443, "analytic1": 3329, "analytic2": null, "analytic3": null, "analytic4": null, "analytic5": null, "analytic6": null, "analytic7": null, "analytic8": null, "analytic9": null, "analytic10": null, "analytic11": null, "analytic12": null, "analytic13": null, "analytic14": null, "analytic15": null, "analytic16": null, "analytic17": null, "analytic18": null, "analytic19": null, "analytic20": null, "summa_turn_debet": 149.0, "summa_turn_credit": 0.0, "summa_total_debet": 149.0, "summa_total_credit": 0.0, "kvd_id": 276, "kosgu_id": 7990, "count_of_entries": 1, "total_entries": 1, "kvd_analytic_id": null, "closing": false, "contract_id": null, "document_type_id": 98769, "past_years_return": false, "currency_id": 96, "product_id": 36464, "unit_id": 28, "fact_accounting": false, "average_accounting": false, "cost": 149.0, "internal": false, "quantity_turn_debet": 1.0, "quantity_turn_credit": 0.0, "quantity_total_debet": 1.0, "quantity_total_credit": 0.0, "journal": "", "transition": false, "sources": "{1474659}", "keys": -2664150223146445388}}
...
2019-09-25 10:38:40,774 - pgdatalogger - INFO - {"_txid": 32884, "_kind": "BEGIN", "_session": "{\"id\": \"762699e883e8#b25ac9c8b638\", \"path\": \"/buh/pay_docs_in_document/unregister\", \"user\": 44, \"is_celery\": false}"} | {"kind": "update", "table": "sufd_pay_doc_in", "key": 29928}
```

Логирование выполняется с использованием логического декодирования 
(logical decoding) и плагина для преобразования данных в json 
([wal2json](https://github.com/eulerto/wal2json) или 
[replisome](https://github.com/GambitResearch/replisome)).

Для логирования информации о сессии используется подход, когда в рамках каждой 
транзакции происходит изменение в специальной таблице  
**commit_trigger.trigger_called**, в которую записывается информация о сессии. 
Поэтому, при выводе журнала изменении, сперва определяется, есть ли в текущей 
транзакции изменения в **commit_trigger.trigger_called** и эти изменения 
выводятся в начале строки для каждой измененной записи в журнале.

Чтобы в каждой транзакции происходили изменения таблицы 
**commit_trigger.trigger_called**, на все таблицы в БД навешиваются триггеры.
Также, при добавлении новой таблицы триггер автоматически навешивается и на нее.

Информация о сессии для записи в таблицу берется из сессионной переменной 
**myvars.session**, в которую надо записать json-строку:
```
SET SESSION myvars.session = '{"url":"https://", "id": "12312", "user": 22}';
``` 
Это значение необходимо выставлять перед началом обработки запроса.
Например, в middleware (подключить PGDataLoggerMiddleware) или перед стартом 
фоновой задачи (подключить celery_handlers).


Настройка PostgreSQL и базы данных
----------------------------------

Согласно инструкции по установке wal2json https://github.com/eulerto/wal2json

1. Установить расширение wal2json

```
    git clone https://github.com/eulerto/wal2json.git
    cd wal2json
    PATH=/path/to/bin/pg_config:$PATH
    USE_PGXS=1 make
    USE_PGXS=1 make install
```

2. Добавить в файл postgresql.conf настройки

```
    wal_level = logical
    max_replication_slots = 3
    max_wal_senders = 1
```
3. Добавить в файл pg_hba.conf настройку

```
    local    replication     bars_web_bb                     trust
```

4. Дать пользователю bars_web_bb права на создание слота логической репликации

```
    alter user bars_web_bb with replication
```

5. Создать функции и триггеры

```
    psql -f ./sql/commit_trigger.sql
```

Настройка сервиса pgdatalogger
------------------------------

Для настойки сервиса журналирования нужно создать конфигурационный файл.
Например:

```
[pgdatalogger]
DATABASE_NAME = web_bb_kalina
DATABASE_USER = web_bb
DATABASE_PASSWORD = web_bb
DATABASE_HOST = 127.0.0.1
DATABASE_PORT = 5432
plugin = wal2json
slot = test_slot1
receiver = Wal2JsonReceiver
consumer = SessionPrinter
pretty-print = True
include_xids = True
include_timestamp = True
logfile=/home/kirov/Projects/pgdatalogger/test.log
log_rotate = D
only_ids = False
```

Где:
* DATABASE_NAME - имя базы данных
* DATABASE_USER - пользователь БД
* DATABASE_PASSWORD - пароль пользователя БД
* DATABASE_HOST - адрес сервера 
* DATABASE_PORT - порт сервера
* dsn - можно указать строку подключения к серверу вместе предыдущих параметров,
postgresql://web_bb:web_bb@127.0.0.1:5432/web_bb_kalina
* slot - имя слота, которые создается для чтения изменений
* receiver - имя класса читателя изменений. сейчас есть Wal2JsonReceiver для 
wal2json, JsonReceiver для replisome
* plugin - имя плагина, который подключен на сервере для Logical Decoding. 
сейчас используется wal2json и replisome
* consumer - имя класса отправителя сообщений. сейчас используется 
SessionPrinter 
* includes - список таблиц со схемами, которые нужно отслеживать; вместо 
includes можно указать параметр includes_file - путь до файла со списком таблиц 
(формат значения аналогичен includes)
* logfile - путь к файлу журнала
* log_rotate - для ротации файлов журналов. D - по дням, H - по часам, 
M - по минутам
* only_ids - признак записи в журнал только ID изменившихся записей, вместо полной записи


Настройка приложения
--------------------

Нужно в settings.py

* добавить в MIDDLEWARE 'pgdatalogger.middleware.PGDataLoggerMiddleware' после 
'ProductionRequestLoggingMiddleware', т.к. данные о сессии включают в себя 
идентификатор запроса 
```
MIDDLEWARE = (
    'django.middleware.common.CommonMiddleware',
    'production_request.middleware.ProductionRequestLoggingMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'pgdatalogger.middleware.PGDataLoggerMiddleware',
)
```
* добавить в CELERY_IMPORTS 'pgdatalogger.celery_handlers' также после 'production_request.celery_handlers'
```
CELERY_IMPORTS = [
    'web_bb.core.asynch.task_handlers',
    'production_request.celery_handlers',
    'pgdatalogger.celery_handlers',
]
``` 